@@ -31,26 +31,6 @@ type universe_decl_option =
   | Cumulative of universe_decl_cumul
   | NonCumulative of universe_decl
 
-[%%if coq = "8.19"] 
-type options = {
-  hoas_holes : hole_mapping option;
-  local : bool option;
-  deprecation : Deprecation.t option;
-  primitive : bool option;
-  failsafe : bool; (* readback is resilient to illformed terms *)
-  ppwidth : int;
-  pp : ppoption;
-  pplevel : Constrexpr.entry_relative_level;
-  using : string option;
-  inline : Declaremods.inline;
-  uinstance : uinstanceoption;
-  universe_decl : universe_decl_option;
-  reversible : bool option;
-  keepunivs : bool option;
-  redflags : RedFlags.reds option;
-  no_tc: bool option;
-}
-[%%else]
 type options = {
   hoas_holes : hole_mapping option;
   local : bool option;
@@ -69,7 +49,6 @@ type options = {
   redflags : RedFlags.reds option;
   no_tc: bool option;
 }
-[%%endif]
 
 type 'a coq_context = {
   (* Elpi representation of the context *)
@@ -131,15 +110,9 @@ val lp2skeleton :
 type coercion_status = Regular | Off | Reversible
 type record_field_spec = { name : Name.t; is_coercion : coercion_status; is_canonical : bool }
 
-[%%if coq = "8.19"]
-val lp2inductive_entry :
-  depth:int -> empty coq_context -> constraints -> State.t -> term ->
-  State.t * ('a list * Entries.mutual_inductive_entry * Univ.ContextSet.t * UnivNames.universe_binders * (bool * record_field_spec list) option * DeclareInd.one_inductive_impls list) * Conversion.extra_goals
-[%%else]
 val lp2inductive_entry :
   depth:int -> empty coq_context -> constraints -> State.t -> term ->
   State.t * (DeclareInd.default_dep_elim list * Entries.mutual_inductive_entry * Univ.ContextSet.t * UnivNames.universe_binders * (bool * record_field_spec list) option * DeclareInd.one_inductive_impls list) * Conversion.extra_goals
-[%%endif]
 
 val inductive_decl2lp :
   depth:int -> empty coq_context -> constraints -> State.t -> (Names.MutInd.t * UVars.Instance.t * (Declarations.mutual_inductive_body * Declarations.one_inductive_body) * (Glob_term.binding_kind list * Glob_term.binding_kind list list)) ->
@@ -182,10 +155,28 @@ val in_elpi_let : Name.t -> term -> term -> term -> term
 val in_elpi_appl : depth:int -> term -> term list -> term
 val in_elpi_match : term -> term -> term list -> term
 val in_elpi_fix : Name.t -> int -> term -> term -> term
-
-
 val in_elpi_name : Name.t -> term
 
+val set_coq : Elpi.API.Ast.Scope.language -> unit
+
+(* gref *)
+val in_elpiast_gref : loc:Ast.Loc.t -> Names.GlobRef.t -> Ast.Term.t
+(* term *)
+val in_elpiast_gr : loc:Ast.Loc.t -> Names.GlobRef.t -> Ast.Term.t
+val in_elpiast_poly_gr : loc:Ast.Loc.t -> Names.GlobRef.t -> Ast.Term.t -> Ast.Term.t
+val in_elpiast_poly_gr_instance : loc:Ast.Loc.t -> Names.GlobRef.t -> UVars.Instance.t -> Ast.Term.t
+val in_elpiast_flex_sort : loc:Ast.Loc.t -> Ast.Term.t -> Ast.Term.t
+val in_elpiast_sort : loc:Ast.Loc.t -> state -> Sorts.t -> Ast.Term.t
+val in_elpiast_prod : loc:Ast.Loc.t -> Names.Name.t -> Ast.Term.t -> Ast.Term.t -> Ast.Term.t
+val in_elpiast_lam : loc:Ast.Loc.t -> Names.Name.t -> Ast.Term.t -> Ast.Term.t -> Ast.Term.t
+val in_elpiast_let : loc:Ast.Loc.t -> Names.Name.t -> ty:Ast.Term.t -> bo:Ast.Term.t -> Ast.Term.t -> Ast.Term.t
+val in_elpiast_appl : loc:Ast.Loc.t -> Ast.Term.t -> Ast.Term.t list -> Ast.Term.t
+val in_elpiast_match : loc:Ast.Loc.t -> Ast.Term.t -> Ast.Term.t -> Ast.Term.t list -> Ast.Term.t
+val in_elpiast_fix : loc:Ast.Loc.t -> Names.Name.t -> int -> Ast.Term.t -> Ast.Term.t -> Ast.Term.t
+val in_elpiast_name : loc:Ast.Loc.t -> Names.Name.t -> Ast.Term.t
+val in_elpiast_decl : loc:Ast.Loc.t -> v:Ast.Term.t -> Names.Name.t -> ty:Ast.Term.t -> Ast.Term.t
+val in_elpiast_def : loc:Ast.Loc.t -> v:Ast.Term.t -> Names.Name.t -> ty:Ast.Term.t -> bo:Ast.Term.t -> Ast.Term.t
+
 val in_coq_name : depth:int -> term -> Name.t
 val is_coq_name : depth:int -> term -> bool
 
@@ -205,11 +196,7 @@ val global_constant_of_globref : Names.GlobRef.t -> global_constant
 val abbreviation : Globnames.abbreviation Conversion.t
 val implicit_kind : Glob_term.binding_kind Conversion.t
 val collect_term_variables : depth:int -> term -> Names.Id.t list
-[%%if coq = "8.19"]
-type pstring = string
-[%%else]
 type pstring = Pstring.t
-[%%endif]
 type primitive_value =
   | Uint63 of Uint63.t
   | Float64 of Float64.t
@@ -217,6 +204,8 @@ type primitive_value =
   | Projection of Projection.t
 val primitive_value : primitive_value Conversion.t
 val in_elpi_primitive : depth:int -> state -> primitive_value -> state * term
+val in_elpiast_primitive : loc:Ast.Loc.t -> primitive_value -> Ast.Term.t 
+
 val uinstance : UVars.Instance.t Conversion.t
 
 val universe_constraint : Univ.univ_constraint Conversion.t
@@ -313,13 +302,16 @@ val update_sigma : State.t -> (Evd.evar_map -> Evd.evar_map) -> State.t
 
 type hyp = { ctx_entry : term; depth : int }
 
-type 'arg tactic_main = Solve of 'arg list | Custom of string
-val goals2query :
-  Evd.evar_map -> Evar.t list -> Elpi.API.Ast.Loc.t -> main:'a tactic_main ->
-  in_elpi_tac_arg:(depth:int -> ?calldepth:int -> 'b coq_context -> hyp list -> Evd.evar_map -> State.t -> 'a -> State.t * term list * Conversion.extra_goals) ->
-  depth:int -> State.t -> State.t * (Elpi.API.Ast.Loc.t * term) * Conversion.extra_goals
-val tclSOLUTION2EVD : eta_contract_solution:bool -> Evd.evar_map -> 'a Elpi.API.Data.solution -> unit Proofview.tactic
-val solution2evd : eta_contract_solution:bool -> Evd.evar_map -> 'a Elpi.API.Data.solution -> Evar.Set.t -> Evd.evar_map * Evar.t list * Evar.t list
+val solvegoals2query :
+  Evd.evar_map -> Evar.t list -> Elpi.API.Ast.Loc.t -> main:'a list ->
+  in_elpi_tac_arg:(base:'base -> depth:int -> ?calldepth:int -> 'b coq_context -> hyp list -> Evd.evar_map -> State.t -> 'a -> State.t * term list * Conversion.extra_goals) ->
+  depth:int -> base:'base -> State.t -> State.t * term * Conversion.extra_goals
+val txtgoals2query :
+  Evd.evar_map -> Evar.t list -> Elpi.API.Ast.Loc.t -> main:string ->
+  depth:int -> base:Elpi.API.Compile.program -> State.t -> State.t * term * Conversion.extra_goals
+
+val tclSOLUTION2EVD : eta_contract_solution:bool -> Evd.evar_map -> Elpi.API.Data.solution -> unit Proofview.tactic
+val solution2evd : eta_contract_solution:bool -> Evd.evar_map -> Elpi.API.Data.solution -> Evar.Set.t -> Evd.evar_map * Evar.t list * Evar.t list
 
 val show_coq_engine : ?with_univs:bool -> State.t -> string
 val show_coq_elpi_engine_mapping : State.t -> string
